/*
 * Copyright (c) 2016-2018, Virtuozzo International GmbH
 *
 * This file is part of disp-helper. It's free software; you can redistribute
 * it and/or modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 *
 * Our contact details: Virtuozzo International GmbH, Vordergasse 59, 8200
 * Schaffhausen, Switzerland.
 */

#include "Client.hpp"
#include <boost/function.hpp>
#include <boost/foreach.hpp>
#include <boost/random/mersenne_twister.hpp>
#include <boost/random/discrete_distribution.hpp>
#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/classification.hpp>
#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/json_parser.hpp>
#include <boost/date_time/c_local_time_adjustor.hpp>
#include <boost/regex.hpp>
#include <boost/algorithm/string/replace.hpp>
#include <fstream>
#include <sys/stat.h>
#include <dlfcn.h>

#include <cstdlib>

namespace Client
{

namespace
{

//public
ScriptBlacklist::ScriptBlacklist(const std::string &scriptsDir) {
	m_scriptsDir = new boost::filesystem::path(scriptsDir);
	loadBlacklist();
}

bool ScriptBlacklist::allowed(fs::directory_entry const& entry_) {
	return m_blacklist->count(entry_.path().string()) == 0;
}

//private
void ScriptBlacklist::loadBlacklist() {
	m_blacklist = new std::set<std::string>();
        boost::property_tree::ptree *root = new boost::property_tree::ptree;
        try {
                boost::property_tree::read_json("/etc/vz/disp_helper.json", *root);
        } catch(...) {
                return;
        }

	try {
		BOOST_FOREACH(boost::property_tree::ptree::value_type &v, root->get_child("noexecute.")) {
			std::string tmp = v.second.data();
			if(tmp[0] == '/') {
				m_blacklist->insert(tmp);
			} else {
				fs::path newPath(m_scriptsDir->string());
				newPath /= tmp;
				m_blacklist->insert(newPath.string());
			}
		}
	} catch(...) {
		return;
	}	
}

static ScriptBlacklist *sb = new ScriptBlacklist(scriptsDir);

std::vector<std::string> GetLogsToCollect()
{
	std::string l(Utils::GlobalSettings().get<std::string>("logs_to_collect", ""));
	if (l.empty())
		return std::vector<std::string>();
	std::vector<std::string> vct;
	boost::split(vct, l, boost::is_any_of(";"), boost::token_compress_on);
	return vct;
}

std::string BuilderFilename()
{
	return "infra_builder.py";
}

bool IsScript(fs::directory_entry const& entry_)
{
	if (!sb->allowed(entry_))
		return false;
	if (!fs::is_regular_file(entry_.status()))
		return false;
	if (!(entry_.status().permissions() & S_IXUSR))
		return false;
	return entry_.path().extension() == ".py";
}

bool IsInfrastructure(fs::directory_entry const& entry_)
{
	return boost::starts_with(entry_.path().filename().string(), "infra_");
}

bool IsPerfomanceScriptPrefix(fs::directory_entry const& entry_)
{
	return boost::starts_with(entry_.path().filename().string(), "perf_");
}

bool IsOrdinaryScript(fs::directory_entry const& entry_)
{
	return IsScript(entry_) && !IsInfrastructure(entry_) && !IsPerfomanceScriptPrefix(entry_);
}

bool IsPerfomanceScript(fs::directory_entry const& entry_)
{
	return IsScript(entry_) && !IsInfrastructure(entry_) && IsPerfomanceScriptPrefix(entry_);
}

void EmptyDummy()
{
}

} // namespace

struct Balancer
{
	Balancer(ba::io_service& service_)
		: m_service(service_),
			m_gen(),
			m_dist(0, 0),
			m_mode(Utils::GlobalSettings().get<std::string>("balancer_mode", "default")),
			m_startDelay(pt::seconds(0)),
			m_actionDelay(pt::seconds(0)),
			m_runNumber(0)
	{
	}

	void Setup(std::time_t lastReport_, fs::path const& scriptsDir_)
	{
		pt::ptime now(pt::second_clock::local_time());
		pt::time_duration reportPeriod(Utils::Datetime::FromString(
					Utils::GlobalSettings().get<std::string>("report_period", "7d")));

		bool changedPeriod = false;
		if (m_runNumber == 0)
		{
			std::string reportPeriodStartupSetting(Utils::GlobalSettings().get<std::string>("report_period_startup", "off"));
			if (reportPeriodStartupSetting != "off")
			{
				pt::time_duration reportPeriodStartup(Utils::Datetime::FromString(reportPeriodStartupSetting));
				if (reportPeriodStartup < reportPeriod)
				{
					reportPeriod = reportPeriodStartup;
					changedPeriod = true;
				}
			}

			m_runNumber += 1;
        }

		pt::ptime lastReport_utc(pt::from_time_t(lastReport_));
		pt::ptime lastReport = boost::date_time::c_local_adjustor<pt::ptime>::utc_to_local(lastReport_utc);

		pt::ptime toSendReport(now + reportPeriod);
		if (!changedPeriod && (now - lastReport) < reportPeriod)
			toSendReport = lastReport + reportPeriod;

		long scriptsNumber(std::count_if(fs::directory_iterator(scriptsDir_),
				fs::directory_iterator(), IsOrdinaryScript));

		// For sufficient delay smoothering before and after sending
		long actionsCount = scriptsNumber + 2;

		if ("off" == m_mode) {
			m_startDelay = toSendReport - now;
			m_actionDelay = pt::seconds(0);
		}

		if ("last_day" == m_mode) {
			m_startDelay = pt::seconds(0);
			if (reportPeriod > pt::hours(24)) {
				m_startDelay = (reportPeriod - pt::hours(24)) - (now - lastReport);
			}
			m_actionDelay = (toSendReport - (now + m_startDelay)) / actionsCount;
		}
		if ("default" == m_mode) {
			m_startDelay = pt::seconds(0);
			m_actionDelay = (toSendReport - now) / actionsCount;
		}

		m_dist = boost::random::uniform_int_distribution<>(
				-m_actionDelay.total_seconds() / 10, m_actionDelay.total_seconds() / 10);
	}

	void StartDelay()
	{
		if ("default" == m_mode)
			return;
		ba::deadline_timer(m_service, m_startDelay).wait();
	}

	void ActionDelay()
	{
		if ("off" == m_mode)
			return;
		pt::time_duration delay = pt::seconds(m_dist(m_gen)) + m_actionDelay;
		ba::deadline_timer(m_service, delay).wait();
	}

private:
	ba::io_service& m_service;
	boost::mt19937 m_gen;
	boost::random::uniform_int_distribution<> m_dist;

	std::string const m_mode;
	pt::time_duration m_startDelay;
	pt::time_duration m_actionDelay;
	int m_runNumber;
};

typedef int (*sender_t)(const char*, const char*, char**);
typedef const char* (*protocol_t)(void);
Collector::Collector(ba::io_service& service_)
	: m_service(service_),
		m_timer(m_service, pt::seconds(1)),
		m_perfomance_timer(m_service, pt::seconds(1)),
		m_scriptsDir(scriptsDir),
		m_preliminaryDir("/vz/tmp/cep/output/preliminary"),
		m_finalResultsDir("/vz/tmp/cep/output/final"),
		m_archive("/vz/tmp/report.tar.gz"),
		m_flagFile("/vz/tmp/cep/.sent"),
		m_balancer(new Balancer(m_service))
{
	fs::directory_iterator last;
	for (fs::directory_iterator it("/usr/lib64/disp-helper-plugins"); it != last; ++it) {
		try {
			if (it->path().extension() != ".so")
				continue;
			::dlerror();
			void* handle = ::dlopen(it->path().string().c_str(), RTLD_LAZY);
			if (NULL == handle) {
				LOG_ERROR_FMT(boost::format("Unable to open shared library %s %s") % it->path().string() % ::dlerror());
				continue;
			}

			protocol_t protocol((protocol_t)::dlsym(handle, "protocol"));
			const char* error(::dlerror());
			if (NULL != error) {
				LOG_ERROR_FMT(boost::format("Unable to load symbols from %s (%s)")
						% it->path().string() % error);
				continue;
			}

			sender_t sender((sender_t)::dlsym(handle, "sender"));
			error = ::dlerror();
			if (NULL != error) {
				LOG_ERROR_FMT(boost::format("Unable to load symbols from %s (%s)")
						% it->path().string() % error);
				continue;
			}
			m_senders[protocol()] = sender;
		} catch (std::exception const& exc) {
			LOG_ERROR_FMT(boost::format("Exception %s") % exc.what());
			exit(EXIT_FAILURE);
		}
	}

	this->distroJson = NULL;
}

bool Collector::ExecuteScripts(boost::function<bool (fs::directory_entry const&)> predicate_,
		boost::function<void ()> delay_) const
{
	fs::directory_iterator last;
	for (fs::directory_iterator it(m_scriptsDir); it != last; ++it) {
		try {
			if (!predicate_(*it))
				continue;

			char **args(NULL);
			try {
				args = static_cast<char**>(::calloc(5, sizeof(char*)));
				args[0] = ::strdup("/usr/bin/python");
				args[1] = ::strdup(fs::absolute(it->path()).string().c_str());
				args[2] = static_cast<char*>(::malloc(128));
				::snprintf(args[2], 128, "%s", m_preliminaryDir.string().c_str());
				args[3] = static_cast<char*>(::malloc(128));
				::snprintf(args[3], 128, "--time=%lu", Utils::Datetime::CurrentTimestamp());
				args[4] = NULL;
				Utils::Base::Instance().IoService().dispatch(boost::bind(Utils::SafeExecute, args, delay_));
				for (size_t i = 0; i < 5; ++i)
					::free(args[i]);
				free(args);
			} catch (std::exception const& exc) {
				if (args) {
					for (size_t i = 0; i < 5; ++i)
						::free(args[i]);
					free(args);
				}
			}
		} catch (std::exception const& e) {
			LOG_ERROR_FMT(boost::format("%s %s") % it->path().filename() % e.what());
			return false;
		}
	}
	return true;
}

void Collector::PrepareFilesystem() const
{
	try {
		boost::system::error_code ec;
		fs::remove_all(m_preliminaryDir, ec);
		fs::remove_all(m_finalResultsDir, ec);
		if (Utils::GlobalSettings().get<std::string>("debug_results", "off") != "on")
			fs::remove(m_archive, ec);
		fs::create_directories(m_preliminaryDir, ec);
		fs::create_directories(m_finalResultsDir, ec);
	} catch (fs::filesystem_error& e) {
		LOG_ERROR_FMT(boost::format("Filesystem error: %s") % e.what());
		exit(EXIT_FAILURE);
	}

	pt::ptime now(pt::second_clock::local_time());
	try {
		if (!fs::exists(m_flagFile))
			Utils::TouchFile(m_flagFile);
		else if (pt::from_time_t(fs::last_write_time(m_flagFile)) > now)
			Utils::TouchFile(m_flagFile);
	} catch (fs::filesystem_error& e) {
		LOG_ERROR_FMT(boost::format("Filesystem error: %s") % e.what());
		exit(EXIT_FAILURE);
	}
}

bool Collector::PrepareResults()
{
	fs::path builder(m_scriptsDir / BuilderFilename());

	char **args(NULL);
	try {
		args = static_cast<char**>(::calloc(6, sizeof(char*)));
		args[0] = ::strdup("/usr/bin/python");
		args[1] = ::strdup(builder.string().c_str());
		args[2] = ::strdup(m_preliminaryDir.string().c_str());
		args[3] = static_cast<char*>(::malloc(128));
		::snprintf(args[3], 128, "--time=%lu", Utils::Datetime::CurrentTimestamp());
		args[4] = ::strdup("/vz/tmp/cep/output/final/result.json");
		args[5] = NULL;
		Utils::SafeExecute(args, EmptyDummy);
	} catch (std::exception const& exc) {
		if (args) {
			for (size_t i = 0; i < 5; ++i)
				::free(args[i]);
			free(args);
		}
	}

	std::vector<std::string> vct(GetLogsToCollect());
	std::string newdate = "XX:XX:XX";
	std::string new_uuid = "XXXXXXXX-XXXX-XXXX-XXXX-XXXXXXXXXXXX";
	std::string new_proc_id = "[XXXXX]";
	boost::regex re_date("\\d*:\\d*:[0-9\\.]*", boost::regex::extended);
	boost::regex re_proc("\\[([^\[\\]]*)\\]");
	boost::regex re_uuid("[0-9a-fA-F]{8}\\-[0-9a-fA-F]{4}\\-[0-9a-fA-F]{4}\\-[0-9a-fA-F]{4}\\-[0-9a-fA-F]{12}");
	for (size_t i = 0; i < vct.size(); ++i) {
		if (!fs::exists(vct[i])) {
			LOG_WARN_FMT(boost::format("Unable to find file %s, please check settings") % vct[i]);
			continue;
		}
		fs::path p(vct[i]);
		try {
			fs::path output_path = m_finalResultsDir / p.filename();
			std::ofstream output(output_path.string().c_str(), std::ofstream::trunc);
			if (!output) {
				LOG_WARN_FMT(boost::format("Unable to open file %s for writing") % output_path.string());
				continue;
			}
			std::ifstream input(p.string().c_str());
			if (!input) {
				LOG_WARN_FMT(boost::format("Unable to open file %s for reading") % p.string());
				output.close();
				continue;
			}
			std::string line;
			std::map<std::string, int> counts;

			while(getline(input, line)) {
				std::string result_date = boost::regex_replace(line, re_date, newdate);
				std::string result_uuid = boost::regex_replace(result_date, re_uuid, new_uuid);
				std::string result = boost::regex_replace(result_uuid, re_proc, new_proc_id);

				if (counts.count(result)) {
					counts[result]++;
				} else {
					counts[result] = 1;
				}
			}

			for (std::map<std::string, int>::iterator iter = counts.begin(); iter != counts.end(); iter++) {
				output << iter->first << " x " << iter->second << std::endl;
			}
			input.close();
			output.close();
		} catch (fs::filesystem_error& e) {
			LOG_WARN_FMT(boost::format("Filesystem error: %s") % e.what());
			continue;
		}
	}

	// TODO use libtar
	std::ostringstream ostr;
	ostr << "/usr/bin/tar -zcf " << m_archive << " --directory=\""
		<< fs::absolute(m_finalResultsDir) << "\" .";
	return !::system(ostr.str().c_str());
}

void Collector::Iteration()
{
	PrepareFilesystem();
	if (!fs::exists(m_flagFile)) {
		LOG_ERROR("Filesystem error: flagFile is missing why we had to create it a moment ago");
		exit(EXIT_FAILURE);
	}
	m_balancer->Setup(fs::last_write_time(m_flagFile), m_scriptsDir);
	m_balancer->StartDelay();
	ExecuteScripts(boost::bind(IsOrdinaryScript, _1), boost::bind(&Balancer::ActionDelay, m_balancer));
	PrepareResults();

	m_balancer->ActionDelay();
	m_service.post(boost::bind(&Collector::SendResults, this));
}

void Collector::PerfomanceIteration()
{
	ExecuteScripts(boost::bind(IsPerfomanceScript, _1), EmptyDummy);
	pt::time_duration frequency(Utils::Datetime::FromString(Utils::GlobalSettings().get<std::string>("frequency_period", "5m")));
	m_perfomance_timer.expires_from_now(frequency);
	m_perfomance_timer.async_wait(boost::bind(&Collector::PerfomanceIteration, this));
}

void Collector::Run()
{
	LOG_INFO_FMT(boost::format("Started collector"));
	Utils::Base::Instance().IoService().dispatch(boost::bind(&Collector::PerfomanceIteration, this));
	Utils::Base::Instance().IoService().dispatch(boost::bind(&Collector::Iteration, this));
}

bool Collector::CheckDistroJson(std::string &protocol)
{
	if (this->distroJson) {
		delete this->distroJson;
	}

	boost::property_tree::ptree *root = new boost::property_tree::ptree;
	try {
		boost::property_tree::read_json("/etc/vz/disp_helper.json", *root);
	} catch(...) {
		// failed to parse
		delete root;
		protocol = Utils::GlobalSettings().get<std::string>("protocol", "CURL");
		return false;
	}

	try {
		protocol = root->get<std::string>("protocol", "");
	} catch(...) {
		protocol = "";
	}

	bool changed = false;
	if (fs::is_regular_file("/etc/hci-release")) {
		if (system("/bin/systemctl is-active vstorage-ui-backend > nul:") == 0) {
			if (protocol != "CURL") {
				root->put("protocol", "CURL");
				protocol = "CURL";
				changed = true;
			}
		} else if (protocol != "FILE") {
			root->put("protocol", "FILE");
			protocol = "FILE";
			changed = true;
		}
	}

	if (changed == true) {
		boost::property_tree::write_json("/etc/vz/disp_helper.json", *root);
		delete root;
		this->distroJson = NULL;
		return true;
	}

	this->distroJson = root;
	protocol = Utils::GlobalSettings().get<std::string>("protocol", "CURL");
	return false;
}

void Collector::SendResults()
{
	std::string protocol;
	bool changed = this->CheckDistroJson(protocol);

	typename senders_t::const_iterator it(m_senders.find(protocol));
	if (m_senders.end() == it) {
		LOG_ERROR_FMT(boost::format("Unable to find sender for '%s' protocol") % protocol);
		::exit(EXIT_FAILURE);
	}

    char *answer(NULL);

    std::string server_default("https://report.virtuozzo.com/post");
    std::string server(Utils::GlobalSettings().get<std::string>("master_server", server_default));
    std::string server_fallback(Utils::GlobalSettings().get<std::string>("master_server_fallback", server));

    int error(it->second(m_archive.string().c_str(), server.c_str(), &answer));
    if (error != 0 || (answer != NULL && strstr(answer, "504 Gateway Time-out") != NULL))
    {
        int error(it->second(m_archive.string().c_str(), server_fallback.c_str(), &answer));
    }

	if (0 == error) {
		Utils::TouchFile(m_flagFile);

		std::vector<std::string> vct(GetLogsToCollect());
		for (size_t i = 0; i < vct.size(); ++i) {
			if (!fs::exists(vct[i]))
				continue;
			std::ofstream ofs;
			ofs.open(vct[i].c_str(), std::ofstream::out | std::ofstream::trunc);
			ofs.close();
		}

		LOG_INFO_FMT(boost::format("Successfully sent report; Server response: %s") % answer);
		m_balancer->ActionDelay();
		m_service.post(boost::bind(&Collector::Iteration, this));
	} else {
		std::string s(Utils::GlobalSettings().get<std::string>("retry_period", "2h"));
		LOG_ERROR_FMT(boost::format("Unable to send. Sender return %d, "
					"retrying send attempt in %s") % error % s);
		pt::time_duration retryPeriod(Utils::Datetime::FromString(s));
		m_timer.expires_from_now(retryPeriod);
		m_timer.async_wait(boost::bind(&Collector::SendResults, this));
	}
	::free(answer);
	if (changed) {
		int dret_code = system("/bin/systemctl restart disp-helper >/dev/null 2>&1");
		if (dret_code != 0) {
			// can't restart disp-helper
		}
	}
}

} // namespace Client
